@@ -117,7 +117,7 @@ fitCyclopsModel <- function(cyclopsData,
                                          })
         }
 
-        if (prior$priorType != "none" &&
+        if (prior$priorType[1] != "none" &&
             is.null(prior$graph) && # TODO Ignore hierarchical models for now
             .cyclopsGetHasIntercept(cyclopsData) &&
             !prior$forceIntercept) {
@@ -559,7 +559,7 @@ createPrior <- function(priorType,
                         neighborhood = NULL,
                         useCrossValidation = FALSE,
                         forceIntercept = FALSE) {
-    validNames = c("none", "laplace","normal", "hierarchical")
+    validNames = c("none", "laplace","normal", "barupdate", "hierarchical")
     stopifnot(priorType %in% validNames)
     if (!is.null(exclude)) {
         if (!inherits(exclude, "character") &&
@@ -574,10 +574,13 @@ createPrior <- function(priorType,
         stop("Prior types and variances have a dimensionality mismatch")
     }
 
-    if (priorType == "none" && useCrossValidation) {
+    if (all(priorType == "none") && useCrossValidation) {
         stop("Cannot perform cross validation with a flat prior")
     }
-    if (priorType == "hierarchical" && missing(graph)) {
+    if (any(priorType == "barupdate") && useCrossValidation) {
+        stop("Cannot perform cross valudation with BAR updates")
+    }
+    if (any(priorType == "hierarchical") && missing(graph)) {
         stop("Must provide a graph for a hierarchical prior")
     }
     if (!is.null(neighborhood)) {
